#include <inttypes.h>
#include <avr\signal.h>
#include <avr\eeprom.h>
#include <avr\pgmspace.h>
#include <avr\delay.h>

#include <stdio.h>
#include "uart.h"




/*

   PD5  = gelbe LED

   PC0  = LED  blau
   PC1  = LED  grn
   PC2  = LED  rot





*/


#define   LED_Y          PD5


#define   LED_B          PC0
#define   LED_G          PC1
#define   LED_R          PC2


#define   LED_MASK       ((1 << LED_B) | (1 << LED_G) | (1 << LED_R))
#define   ROW_MASK       ((1 << PC3) | (1 << PC4) | (1 << PC5) | (1 << PC6))

#define   OCR1A_VALUE                   10646     // 1500Hz
#define   KEY_REPEAT_START                240
#define   KEY_REPEAT_NEXT                  60
#define   SENSOR_THRESHOLD_LOW             55
#define   SENSOR_THRESHOLD_HIGH            70
#define   SENSOR_THRESHOLD_ON               8

volatile uint8_t  SensorCalibrate;                // Kalibrierungszhler, bei fertiger Kalibrierung == 0xFF
volatile uint8_t  SensorState[5];
volatile uint8_t  SensorBurst[5];                 // Anzahl der Burst beim Laden der Sensortasten
volatile uint8_t  SensorSample[5];                // gesamplete Sensorwerte
volatile uint8_t  SensorOffset[5];                // Offset auf Null der Keysamples, entfernt leichte kapazitve Unterschiede zwischen den Sensortasten
volatile uint8_t  SensorCount[5];

volatile uint8_t  KeyState;                       // actual active keys
volatile uint8_t  KeyPressed;                     // pressed keys
volatile uint8_t  KeyRepeated;                    // repeated keys

volatile uint16_t DisplayRAM[15][5];
volatile uint8_t  DisplayRow = 15;

volatile  int8_t  Temp1;
volatile  int8_t  Temp2;

volatile uint8_t  Changed;


SIGNAL(SIG_ADC) {

    static  uint8_t dr; // dummy ADC read to safe change ARef
    static  int32_t t1; // temp1
    static  int32_t t2; // temp2
    static uint16_t tc; // temp counter for average

    ADCSRA &= ~(1 << ADEN);

    if (dr) {
      dr--;
    } else {
      int16_t v = ADC;
      uint8_t i = DisplayRow;
      if (i == 5) {        // Temp1
        t1 += v;
      } else if (i == 6) { // Temp2
        t2 += v;
        if (++tc >= 504) { // 504 + 100Hz = ca. alle 5 Sekunden
          t1 -= (int32_t)(2.7315 * 1024 * 100);   // t = (ADC * 5.04V / 1024 - 2.7315) * 100
          t2 -= (int32_t)(2.7315 * 1024 * 100);   // t = (5.04 * 100 * ADC - (2.7315 * 1024 * 100)) / 1024
          Temp1 = t1 / 1024;                      // t = (504 * ADC - 279706) / 1024
          Temp2 = t2 / 1024;                      // t == Grad Celcius
          t1 = 0;
          t2 = 0;
          tc = 0;
        }
      } else if (i == 7) { // Temp3 ?

        ADMUX  = (1 << REFS1) | (1 << REFS0);
        ADCSRA = (0 << ADSC) | (1 << ADEN) | (1 << ADIF) | (1 << ADIE) | (1 << ADPS2) | (0 << ADPS1) | (0 << ADPS0);
        dr++;
      } else if (i == 13) { // light

      } else if (i == 14) { // ??

        ADMUX  = (1 << REFS1) | (1 << REFS0);
        ADCSRA = (0 << ADSC) | (1 << ADEN) | (1 << ADIF) | (1 << ADIE) | (1 << ADPS2) | (0 << ADPS1) | (0 << ADPS0);
        dr++;
      } else {
        i &= 0x07;
        if (v >= 0x200) v |= 0xFC00; // 10 Bit signed in 16 bit signed
        v  = 512 -v;
        v /= 2;
        uint8_t calibrate = SensorCalibrate;
        if (calibrate != 0xFF) {
          uint8_t burst = SensorBurst[i];
          if (v < 64) {
            burst++;
            calibrate = 0;
          } else if (v > 128) {
            burst--;
            calibrate = 0;
          } else {
            v += SensorOffset[i];
            v /= 2;
            SensorOffset[i] = v;
            calibrate++;
          }
          SensorBurst[i] = burst;
          SensorCalibrate = calibrate;
        } else {
          v -= SensorOffset[i];
          v += SensorSample[i];
          v /= 2;

          static uint8_t SensorRecalibrate;
          uint8_t recalibrate = SensorRecalibrate;
          if ((v < 0) | (v > 255)) { // check limits
            if (v < 0) v = 0;
              else v = 255;
            if (recalibrate >= 252) { // must be recalibrated ?
              recalibrate = 0;
              SensorCalibrate = 0;
            } else {
              recalibrate += 4;
            }
          } else if (recalibrate) {
            recalibrate--;
          }
          SensorRecalibrate = recalibrate;

          uint8_t sample = v;
          uint8_t state = SensorState[i];
          uint8_t count = SensorCount[i];

          if (state == 0) { // wait for key press
            if (sample > SENSOR_THRESHOLD_LOW) {
              state = 1;
              count = 0;
            }
          } else if (state == 1) { // counting up key press, measure speed of keypress
            if (sample > SensorSample[i]) {
              if (++count >= 4) {
                if (sample > SENSOR_THRESHOLD_HIGH) state = 2;
                  else state = 4;
                count = 0;
              }
            } else {
              state = 4;
              count = 0;
            }
          } else if (state == 2) { // wait delay for key press, debounce
            if (sample > SENSOR_THRESHOLD_HIGH) {
              if (++count >= SENSOR_THRESHOLD_ON) {
                state = 3;
                count = 0;
              }
            } else {
              state = 4;
              count = 0;
            }
          } else if (state == 3) { // key on
            if (sample < SENSOR_THRESHOLD_HIGH) {
              if (++count >= 4) {
                state = 4;
                count = 0;
              }
            } else {
              count = 0;
            }
          } else if (state == 4) { // wait for stable key off
            if (sample < SENSOR_THRESHOLD_LOW) {
              if (++count >= 4) {
                state = 0;
                count = 0;
              }
            } else {
              count = 0;
            }
          }
          SensorState[i] = state;
          SensorCount[i] = count;
          SensorSample[i] = sample;
        }

        if (i == 4) {
          ADMUX  = (0 << REFS1) | (1 << REFS0);
          ADCSRA = (0 << ADSC) | (1 << ADEN) | (1 << ADIF) | (1 << ADIE) | (1 << ADPS2) | (0 << ADPS1) | (0 << ADPS0);
          dr++;
         }
      }
    }
}

SIGNAL(SIG_OUTPUT_COMPARE1B) {

// Display is off, Latch is active, select next display row
    uint8_t i = DisplayRow;
    if (++i >= 15) i = 0;  // 100Hz
    DisplayRow = i;
    i <<= 3;
    PORTC = (PORTC & ~ROW_MASK) | i;            
}

SIGNAL(SIG_OVERFLOW1) {

    uint8_t i = DisplayRow;
    uint8_t v = i +1;
    if (v >= 15) v = 0;
    volatile register uint8_t* p asm("r30") = (uint8_t*)&DisplayRAM[v];
    v = p[1];
    SPDR = v;
    v = p[0];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    v = p[3];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    v = p[2];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    v = p[5];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    v = p[4];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    v = p[7];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    v = p[6];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    v = p[9];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    v = p[8];
    while (!(SPSR & (1 << SPIF)));
    SPDR = v;
    while (!(SPSR & (1 << SPIF)));

    if ((i == 5) | (i == 13)) { // update keystate

    }

    if (i == 5) {
      Changed++;
      ADMUX  = (0 << REFS1) | (1 << REFS0) | 0x02;    // temp1
    } else if (i == 6) {
      ADMUX  = (0 << REFS1) | (1 << REFS0) | 0x03;    // temp2
    } else if (i == 7) {
      ADMUX  = (0 << REFS1) | (1 << REFS0);           // temp3
    } else if (i == 13) {
      Changed++;
      ADMUX  = (0 << REFS1) | (1 << REFS0) | 0x00;    // light
    } else if (i == 14) {
      ADMUX  = (0 << REFS1) | (1 << REFS0);           // ?
    } else {
      uint8_t d1 = DDRF & ~(1 << PF0);
      uint8_t d2 = d1 | (1 << PF0);

      DDRF = d2;  // discharge Cs
      _delay_loop_1(11);
      DDRF = d1;

      i &= 0x07;
      uint8_t burst = SensorBurst[i];
      uint8_t p1 = PORTF & ~(1 << PF0);
      uint8_t p2 = p1 | (1 << PF0);

      while (1) {
        DDRE  &= ~(1 << PE0);
        PORTF  = p2;
        DDRF   = d2;
        DDRF   = d1;
        PORTF  = p1;
        DDRE  |=  (1 << PE0);
        if (!(--burst)) break;
      }
      ADMUX  = (1 << REFS1) | (1 << REFS0) | 0x09;
    }
    ADCSRA  = (0 << ADSC) | (1 << ADEN) | (1 << ADIF) | (1 << ADIE) | (1 << ADPS2) | (0 << ADPS1) | (0 << ADPS0);
}


int main(void) {

    DDRA  = 0x00;
    DDRB  = 0x00;
    DDRC  = 0x00;
    DDRD  = 0x00;
    DDRE  = 0x00;
    DDRF  = 0x00;
    DDRG  = 0x00;
    PORTA = 0xFF;
    PORTB = 0xFF;
    PORTC = 0xFF;
    PORTD = 0xFF;
    PORTE = 0xFF;
    PORTF = 0xF0;
    PORTG = 0xFF;

    DDRD   = (1 << LED_Y) | (1 << PD4);
    PORTD  = (0 << LED_Y) | (0 << PD4);

// ADC/AC init
    SFIOR  = 0;
    ACSR   = (1 << ACD) | (1 << ACBG);
    ADCSRA = 0;
    ADMUX  = (1 << REFS1) | (1 << REFS0) | 0x09;

    DDRE  &=  (1 << PE0);  // Cs_GND
    PORTE &= ~(1 << PE0);
    DDRF  &= ~(1 << PF0);  // Cs_V+
    PORTF &= ~(1 << PF0);
    DDRF  &= ~(1 << PF1);  // Vref
    PORTF &= ~(1 << PF1);

// PORT Init
    DDRC   = LED_MASK | ROW_MASK;
    PORTC  = LED_MASK;

    DDRB   = (1 << PB6) | (1 << PB2) | (1 << PB1) | (1 << PB0);

// SPI, 8Mhz, LSB first
    SPCR   = (1 << SPE) | (1 << DORD) | (1 << MSTR) | (0 << CPOL) | (0 << CPHA) | (0 << SPR1) | (0 << SPR0);
    SPSR   = (1 << SPI2X);

// Timer1, Prescaler = 1, OCR1A = TOP = 10647 -> 1500Hz, OCR1B = BLANK = set on compare match, clear on TOP, WGM = Fast PWM
// SIG_OUTPUT_COMPARE1B macht Multiplexing, einlesen der Sensortasten und ADC Werte
// OC1B = PB6 auf Ausgang = BLANK


    TCCR1A = (1 << COM1B1) | (1 << COM1B0) | (1 << WGM11) | (1 << WGM10);
    TCCR1B = (1 << WGM13) | (1 << WGM12) | (0 << CS12) | (0 << CS11) | (1 << CS10);
    TCCR1C = 0;
    ICR1   = 0;
    OCR1A  = OCR1A_VALUE;
    OCR1B  = 5000;
    OCR1C  = 0;

// Timer0 asynchron

    TCCR0  = (1 << CS02) | (0 << CS01) | (0 << CS00);
    ASSR   = (1 << AS0);


    TIMSK  = (1 << OCIE1B) | (1 << TOIE1);
    ETIMSK = 0;

    MCUCR = (1 << SE) | (0 << SM2) | (0 << SM1) | (0 << SM0);
    MCUCSR = (1 << JTD);

    asm("sei");

    uart1_Init(B115200);
//    fdevopen(uart1_Put, uart1_Get, 0);
//    printf("Hallo Test\n");

    PORTD ^= (1 << LED_Y);
    while (1) {
      asm("sleep");

      if (Changed) {
        uart1_Put(0xAA);
        uart1_Put(0x55);
        uart1_Write(&SensorSample, 5);
        uart1_Write(&SensorBurst, 5);
        uart1_Write(&SensorState, 5);
        uart1_Write(&Temp1, 1);
        uart1_Write(&Temp2, 1);
        uart1_Put(0xCC);
        uart1_Put(0xDD);
        Changed = 0;
      }
/*
      if (uart1_RxCount() >= 4) {
        if (uart1_Get() == 0xAA) {
          if (uart1_Get() == 0x55) {
            uart1_Get();
            uart1_Get();
          }
        }
      } */
    }
}


